let state = "idle";
const getPropagationState = () => state;
const eventListeners = {
    freezeEnd: new Set(),
    propagatingEnd: new Set(),
};
const removeEventListener = (event, callback) => {
    eventListeners[event].delete(callback);
};
const addEventListener = (event, callback) => {
    eventListeners[event].add(callback);
};
const runOncePerTick = (callback) => {
    removeEventListener("freezeEnd", callback);
    removeEventListener("propagatingEnd", callback);
    if (state === "frozen") {
        addEventListener("freezeEnd", callback);
    }
    else if (state === "propagating") {
        addEventListener("propagatingEnd", callback);
    }
    else {
        callback();
    }
};
let activeRoot = null;
const rootChildren = new WeakMap();
const rootOfStream = new WeakMap();
/**
 * The computations that will need to rerun next tick
 */
let toRun = new Set();
/**
 * The direct dependencies of a stream, used in computeDependents
 * to compute all descendant dependencies at tick time
 */
let dependents = new WeakMap();
/**
 * A computation can have n cleanups, we store them in a set for each stream
 * and clear the list every tick
 */
let cleanups = new Map();
/**
 * Track the computation call stack so we know which computations
 * depend on other streams when read in their context.
 *
 * Also used to store the computations parents at definition time.
 */
let active = [];
/**
 * When a stream updates, we do not immediately update the
 * value, instead we set stream.next.  We then queue up
 * the stream value to update at the end of the tick
 * by putting those streams in a queue.
 */
let streamsToResolve = new Set();
/**
 * If a write happens when a propagation is already running
 * we store the set invocation in a list so we can trigger
 * the tick once the current tick completes.
 */
let nextTicks = [];
/**
 * The parents of a computation, used to prevent children
 * computation streams being recomputed if the outer
 * computation will already be running it for free.
 *
 * Without this your inner most computations will
 * update ^n number of nested computations, not fun!
 */
let parents = new WeakMap();
let children = new WeakMap();
const ids = new WeakMap();
const accessorStreams = new WeakMap();
const streamAccessors = new WeakMap();
const randomId = () => Math.random().toString(15);
const record = (stream, x) => {
    ids.set(x, randomId());
    accessorStreams.set(x, stream);
    streamAccessors.set(stream, x);
    return x;
};
class StreamError extends Error {
}
class CleanupWithoutComputationContext extends StreamError {
}
class Conflict extends StreamError {
}
class RunawayTicks extends StreamError {
}
const MAX_TICKS = 100000;
let runawayTicks = 0;
/**
 * For tests and debugging.
 */
const stats = {
    ticks: 0,
    computations: {
        evaluated: 0,
    },
};
// Typed version of https://www.npmjs.com/package/xet
// by @barneycarroll
function xet(map, key, fn) {
    if (map.has(key)) {
        return map.get(key);
    }
    const value = fn(key, map);
    map.set(key, value);
    return value;
}
/**
 * When some data changes, direct dependencies need to
 * be updated.  But if those dependencies update, the
 * dependendies of those dependenies need to update.
 *
 * When we know we're about to tick, we compute
 * that recursive structure into a flat list before
 * the tick starts.
 *
 * I prefer it this way, but it also works well with
 * S' atomic state update model.  Only the dependencies
 * at the time the tick started matter.  If by the time
 * the tick is half way complete some new dependency
 * arises, that is a job for another tick.
 *
 * @param stream
 */
function computeDependents(stream) {
    let stack = [];
    stack.push(...xet(dependents, stream, () => new Set()));
    dependents.get(stream)?.clear();
    while (stack.length) {
        let x = stack.shift();
        if (x.tag !== "GeneratorComputation") {
            stack.push(...xet(dependents, x, () => new Set()));
            // already disposed, cheaper to do it here
            if (parents.get(x) == null) {
                // as it doesn't run, it won't
                // be added as a child ever again
                continue;
            }
        }
        toRun.add(x);
    }
}
const strictEquality = (a, b) => a === b;
function data(value, equality = strictEquality) {
    const stream = {
        type: "Stream",
        tag: "Data",
        next: value,
        value,
    };
    let accessor = (...args) => {
        if (args.length) {
            if (state === "propagating") {
                nextTicks.push(() => accessor(...args));
                return stream.value;
            }
            let nextVal;
            if (typeof args[0] === "function") {
                nextVal = args[0](stream.value);
            }
            else {
                nextVal = args[0];
            }
            if (streamsToResolve.has(stream) && !equality(nextVal, stream.next)) {
                throw new Conflict();
            }
            else {
                stream.next = nextVal;
                streamsToResolve.add(stream);
            }
            computeDependents(stream);
            if (state === "idle") {
                tick();
            }
        }
        else if (active[0]) {
            xet(dependents, stream, () => new Set()).add(active[0]);
            return stream.next;
        }
        else {
            return stream.value;
        }
    };
    return record(stream, accessor);
}
const SKIP = {};
function computation(fn, seed) {
    const stream = {
        type: "Stream",
        tag: "SyncComputation",
        value: seed,
        next: seed,
        compute() {
            active.unshift(stream);
            const out = fn(stream.value);
            active.shift();
            if (out === SKIP) {
                return stream.value;
            }
            stream.next = out;
            if (state === "idle") {
                stream.value = stream.next;
            }
            else {
                streamsToResolve.add(stream);
            }
        },
    };
    // whatever active was when this was defined
    // is our parents
    parents.set(stream, new Set(active));
    for (let x of active) {
        xet(children, x, () => new Set()).add(stream);
    }
    //   if (activeRoot === null) {
    //     throw new ComputationWithoutRoot();
    //   }
    // so we can dispose this when the root is
    // disposed
    xet(rootChildren, activeRoot, () => new Set()).add(stream);
    if (activeRoot != null) {
        rootOfStream.set(stream, activeRoot);
    }
    // for when initial run of a computation triggers a write
    // which should behave as if time is frozen
    if (state === "idle") {
        state = "propagating";
        stream.compute();
        state = "idle";
        // some nested writes happened
        if (nextTicks.length > 0) {
            tick();
        }
        else {
            for (let s of streamsToResolve) {
                s.value = s.next;
            }
            streamsToResolve.clear();
        }
    }
    else {
        stream.compute();
    }
    return record(stream, () => {
        if (active[0]) {
            xet(dependents, stream, () => new Set()).add(active[0]);
            return stream.next;
        }
        else {
            return stream.value;
        }
    });
}
function generator(_fn, seed) {
    let iteration;
    let it;
    let fn = _fn;
    let sentinel = {};
    const stream = {
        type: "Stream",
        tag: "GeneratorComputation",
        value: seed,
        compute() {
            if (iteration) {
                it.return(sentinel);
            }
            it = fn(stream.value);
            iteration = iterate(it);
            iteration.catch((err) => {
                it?.throw(err);
            });
            iteration.then((maybeSentinel) => {
                iteration = null;
                it = null;
                if (maybeSentinel === sentinel) {
                    return;
                }
                computeDependents(stream);
                if (state == "idle") {
                    streamsToResolve.add(stream);
                    tick();
                }
            });
        },
    };
    async function iterate(it) {
        let value, done, next;
        do {
            active.unshift(stream);
            void ({ done, value } = it.next(next));
            active.shift();
            if (value && "then" in Object(value)) {
                value = await value;
            }
            next = value;
        } while (!done);
        stream.next = value;
        streamsToResolve.add(stream);
        return value;
    }
    // whatever active was when this was defined
    // is our parents
    parents.set(stream, new Set(active));
    for (let x of active) {
        xet(children, x, () => new Set()).add(stream);
    }
    //   if (activeRoot === null) {
    //     throw new ComputationWithoutRoot();
    //   }
    // so we can dispose this when the root is
    // disposed
    xet(rootChildren, activeRoot, () => new Set()).add(stream);
    if (activeRoot != null) {
        rootOfStream.set(stream, activeRoot);
    }
    stream.compute();
    return record(stream, () => {
        if (active[0]) {
            xet(dependents, stream, () => new Set()).add(active[0]);
            return stream.next;
        }
        return stream.value;
    });
}
function freeze(f) {
    if (state == "frozen") {
        f();
        return;
    }
    // If called within a computation, the system is already frozen, so freeze is inert.
    if (state == "propagating") {
        f();
        return;
    }
    let oldState = state;
    state = "frozen";
    f();
    state = oldState;
    for (let callback of eventListeners.freezeEnd) {
        callback();
    }
    tick();
}
function cleanup(f) {
    if (active[0]) {
        xet(cleanups, active[0], () => new Set()).add(f);
    }
    else {
        throw new CleanupWithoutComputationContext();
    }
}
function root(f) {
    let root = {
        type: "Root",
        tag: "Root",
    };
    let oldActiveRoot = activeRoot;
    activeRoot = root;
    let oldActive = active;
    active = [];
    let out = f(() => {
        for (let x of rootChildren.get(root) ?? []) {
            for (let cleanup of cleanups.get(x) ?? []) {
                cleanup();
            }
            dependents.delete(x);
            toRun.delete(x);
            cleanups.delete(x);
            parents.delete(x);
        }
    });
    activeRoot = oldActiveRoot;
    active = oldActive;
    return out;
}
function tick() {
    if (state === "propagating" || state === "frozen")
        return;
    stats.ticks++;
    state = "propagating";
    let oldToRun = new Set(toRun);
    toRun.clear();
    let allChildren = new Set();
    for (let x of oldToRun) {
        for (let child of children.get(x) ?? []) {
            // record the child exists
            allChildren.add(child);
            // add to the tick so we can run
            // the clean up fns (at most once)
            oldToRun.add(child);
            parents.delete(child);
            streamsToResolve.delete(child);
            // should we delete the dependents too?
            dependents.delete(x);
        }
        // start again
        children.delete(x);
    }
    for (let f of oldToRun) {
        for (let cleanupFn of xet(cleanups, f, () => new Set())) {
            cleanupFn();
            cleanups.get(f).delete(cleanupFn);
        }
        // if is a child, we only need to clean up
        if (allChildren.has(f)) {
            continue;
        }
        else {
            let oldActive = active;
            active = [...(parents.get(f) ?? [])];
            let oldActiveRoot = activeRoot;
            activeRoot = rootOfStream.get(f) ?? null;
            f.compute();
            active = oldActive;
            activeRoot = oldActiveRoot;
            stats.computations.evaluated++;
        }
    }
    for (let s of streamsToResolve) {
        s.value = s.next;
    }
    streamsToResolve.clear();
    state = "idle";
    if (runawayTicks > 0)
        return;
    // while processing a tick, nextTicks
    // may expand, so we need to recursively
    // update
    // we use a while loop to prevent stack
    // overflow, the above early exit detects
    // when our tick is nested and allows the
    // parent tick to handle resuming the propagation
    while (nextTicks.length > 0) {
        runawayTicks++;
        freeze(() => {
            let xs = nextTicks.slice();
            nextTicks.length = 0;
            for (let next of xs) {
                if (runawayTicks > MAX_TICKS) {
                    throw new RunawayTicks();
                }
                next();
            }
        });
        if (nextTicks.length == 0) {
            for (let callback of eventListeners.propagatingEnd) {
                callback();
            }
        }
    }
    runawayTicks = 0;
}
function sample(signal) {
    let oldActive = active;
    active = [];
    let value = signal();
    active = oldActive;
    return value;
}
function id(s) {
    return ids.get(s);
}

export { CleanupWithoutComputationContext, Conflict, MAX_TICKS, RunawayTicks, SKIP, StreamError, accessorStreams, addEventListener, children, cleanup, computation, data, dependents, freeze, generator, getPropagationState, id, ids, parents, removeEventListener, root, runOncePerTick, runawayTicks, sample, stats, streamAccessors, streamsToResolve, tick, toRun, xet };
//# sourceMappingURL=index.js.map
